package dsekercioglu.mega.wfGun;

import dsekercioglu.mega.wfGun.gun.AntiAdaptiveGun;
import dsekercioglu.mega.wfGun.gun.AbstractPredictor;
import dsekercioglu.mega.wfGun.gun.AntiAdaptiveGun;
import dsekercioglu.mega.wfGun.gun.MainGun;
import dsekercioglu.mega.wfGun.precise.WallSmoothingMEACalculator;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import robocode.AdvancedRobot;
import robocode.BulletHitBulletEvent;
import robocode.BulletHitEvent;
import robocode.RoundEndedEvent;
import robocode.ScannedRobotEvent;
import static robocode.util.Utils.normalRelativeAngle;

public class Fangs {

    AdvancedRobot a;

    public final int BINS = 51;

    public final ArrayList<AbstractPredictor> PREDICTORS = new ArrayList<>();
    public final ArrayList<Double> SCORES = new ArrayList<>();
    public final double SCORE_DECAY = 0.995;
    public final double SURVIVALIST_TRESHOLD = 0;

    ArrayList<Wave> waves = new ArrayList<>();

    FirePowerChooser firePowerChooser = new FirePowerChooser();

    WallSmoothingMEACalculator WSMEAC;

    BattleInfo battleInfo = new BattleInfo(this);

    public double shotsTaken = 0;
    public double bulletHitBullet = 0;
    public double bulletsHit = 0;

    Wave lastWave;

    Point2D.Double nextPosition;

    ArrayList<Point2D.Double> fireLocations = new ArrayList<>();
    ArrayList<Double> fireAngles = new ArrayList<>();
    ArrayList<Double> maxDev = new ArrayList<>();

    final boolean TC;

    public Fangs(AdvancedRobot a, boolean TC) {
        this.a = a;
        this.TC = TC;
        PREDICTORS.add(new MainGun(BINS));
        PREDICTORS.add(new AntiAdaptiveGun(BINS));
        for (int i = 0; i < PREDICTORS.size(); i++) {
            SCORES.add(0.0);
        }
    }

    public void run() {
        waves = new ArrayList<>();
        fireAngles = new ArrayList<>();
        battleInfo = new BattleInfo(this);
        battleInfo.run();
        WSMEAC = new WallSmoothingMEACalculator(battleInfo.battleFieldWidth, battleInfo.battleFieldHeight, 0.01);
    }

    public void setNextPosition(Point2D.Double nextPosition) {
        this.nextPosition = (Point2D.Double) nextPosition.clone();
    }

    public void onScannedRobot(ScannedRobotEvent e) {
        battleInfo.onScannedRobot(e);
        updateWaves();
        double gunHeat = a.getGunHeat();
        int lateralDir = battleInfo.lateralDirection;
        double firePower = firePowerChooser.firePower();
        double bulletSpeed = (20 - 3 * firePower);
        Wave w = new Wave(firePower, nextPosition, battleInfo.enemyLocation, lateralDir, false);
        double[] data = {Math.abs(battleInfo.lateralVelocity) / 8,
            battleInfo.advancingVelocity / 16 + 0.5,
            battleInfo.distance / bulletSpeed / 91,
            Math.min(battleInfo.getMEA(1), w.maxEscapeAngle) / w.maxEscapeAngle,
            Math.min(battleInfo.getMEA(-1), w.maxEscapeAngle) / w.maxEscapeAngle,
            GunUtils.limit(0, (battleInfo.lateralAcceleration + 2) / 3, 1),
            1.0 / (1 + battleInfo.timeSinceDeceleration * 0.1),
            battleInfo.getLateralDistanceLastX(10) / 8,
            shotsTaken};
        w.setData(data);
        w.setBins(predictBins(data));
        WSMEAC.calculateEscapeAngle(battleInfo.enemyLocation, nextPosition, bulletSpeed, bulletSpeed);
        Point2D.Double p1 = WSMEAC.getEscapePosition(-1);
        Point2D.Double p2 = WSMEAC.getEscapePosition(1);
        int b1 = w.getBin(p1);
        int b2 = w.getBin(p2);
        if (a.getGunHeat() <= 0.4 && battleInfo.botEnergy - firePower >= SURVIVALIST_TRESHOLD) {
            double fireAngle = w.getFireAngle(Math.min(b1, b2), Math.max(b1, b2));
            a.setTurnGunRightRadians(normalRelativeAngle(fireAngle - a.getGunHeadingRadians()) + (bulletHitBullet / shotsTaken > 0.3 ? Math.random() * 9 / battleInfo.distance : 0));
            if (Math.abs(a.getGunTurnRemainingRadians()) < Math.asin(18 / (battleInfo.distance - 18)) && gunHeat == 0) {
                w.real = true;
                ArrayList<Object> guessFactors = new ArrayList<>();
                for (int i = 0; i < PREDICTORS.size(); i++) {
                    guessFactors.add(GunUtils.getMostVisitedBin(PREDICTORS.get(i).predict(data)));
                }
                w.setInformation(guessFactors);

                a.setFire(firePower);
                lastWave = w;

                fireAngles.add(0, fireAngle);
                fireLocations.add(0, w.fireLocation);
                maxDev.add(0, Math.asin(18 / (battleInfo.distance - 18)));

                shotsTaken++;
            }
        } else {
            a.setTurnGunRightRadians(normalRelativeAngle(w.absBearing - a.getGunHeadingRadians()));
        }
        if (battleInfo.botEnergy - firePower >= 0) {
            waves.add(w);
        }
    }

    public void updateWaves() {
        for (int i = 0; i < waves.size(); i++) {
            Wave w = waves.get(i);
            boolean remove = w.update(battleInfo.enemyLocation);
            if (remove) {
                int bin = w.getBin(battleInfo.enemyLocation);
                for (int j = 0; j < PREDICTORS.size(); j++) {
                    PREDICTORS.get(j).wavePassed(w.data, bin, w.real);
                    if (w.real) {
                        //double scoreIncreasement = 1 / (Math.pow((Integer) w.information.get(j) - bin, 2) + 1);
                        double scoreIncreasement = (Math.abs((Integer) w.information.get(j) - bin) * w.binWidth) < Math.asin(18 / (w.fireLocation.distance(battleInfo.enemyLocation) - 18)) ? 1 : 0;
                        SCORES.set(j, SCORES.get(j) * SCORE_DECAY + scoreIncreasement * (1 - SCORE_DECAY));
                    }
                }
                waves.remove(i);
                i--;
            }
        }
    }

    public double[] predictBins(double[] data) {
        int bestIndex = 0;
        double bestScore = SCORES.get(0);
        for (int i = 1; i < PREDICTORS.size(); i++) {
            if (SCORES.get(i) > bestScore) {
                bestScore = SCORES.get(i);
                bestIndex = i;
            }
        }
        return PREDICTORS.get(bestIndex).predict(data);
//        double[] sumBins = new double[BINS];
//        for (int i = 0; i < PREDICTORS.size(); i++) {
//            double[] currentBins = PREDICTORS.get(i).predict(data);
//            double score = SCORES.get(i);
//            for (int j = 0; j < sumBins.length; j++) {
//                sumBins[j] += currentBins[j];
//            }
//        }
//        return sumBins;
    }

    public class BattleInfo {

        Point2D.Double location = new Point2D.Double();
        Point2D.Double enemyLocation = new Point2D.Double();

        double battleFieldWidth;
        double battleFieldHeight;

        public double velocity;
        public double heading;
        public double distance;
        public double bearing;
        public double botEnergy;
        public double enemyEnergy;

        public double relativeAngle;
        public double lateralVelocity;
        public double advancingVelocity;
        public double lateralAcceleration;

        public int lateralDirection;

        private ArrayList<Double> lateralVelocityList = new ArrayList<>();

        public int timeSinceDeceleration;

        public double botHeading;

        public final Fangs g;

        public BattleInfo(Fangs g) {
            this.g = g;
        }

        public void run() {
            battleFieldWidth = a.getBattleFieldWidth();
            battleFieldHeight = a.getBattleFieldHeight();
        }

        public void onScannedRobot(ScannedRobotEvent e) {
            timeSinceDeceleration++;

            location.setLocation(a.getX(), a.getY());
            botHeading = a.getHeadingRadians();
            botEnergy = a.getEnergy();
            velocity = e.getVelocity();
            heading = e.getHeadingRadians();
            bearing = e.getBearingRadians();
            distance = e.getDistance();
            enemyLocation = GunUtils.project(location, a.getHeadingRadians() + bearing, distance);
            enemyEnergy = e.getEnergy();
            relativeAngle = heading - (GunUtils.absoluteBearing(location, enemyLocation));
            double newLateralVelocity = velocity * Math.sin(relativeAngle);
            lateralAcceleration = Math.abs(newLateralVelocity) - Math.abs(lateralVelocity);
            if (lateralAcceleration < 0) {
                timeSinceDeceleration = 0;
            }
            lateralVelocity = newLateralVelocity;
            lateralDirection = lateralVelocity >= 0 ? 1 : -1;
            advancingVelocity = velocity * -Math.cos(relativeAngle);

            lateralVelocityList.add(lateralVelocity);
        }

        public double getMEA(int direction) {
            double turnPerTick = 8 / distance;
            double maxIteration = Math.PI * 2 / turnPerTick;
            double angle = botHeading + bearing;
            double mea = 0;
            Point2D.Double predictedPosition = GunUtils.project(location, angle, distance);
            int iteration = 0;
            while (GunUtils.distanceToWall(predictedPosition.x,
                    predictedPosition.y,
                    battleFieldWidth,
                    battleFieldHeight) >= 18) {
                angle += turnPerTick * lateralDirection * direction;
                mea += turnPerTick;
                predictedPosition = GunUtils.project(location, angle, distance);
                if ((++iteration) > maxIteration) {
                    break;
                }
            }
            return mea;
        }

        public double getLateralDistanceLastX(int x) {
            double lateralDistance = 0;
            int startIndex = Math.max(0, lateralVelocityList.size() - x);
            for (int i = startIndex; i < lateralVelocityList.size(); i++) {
                lateralDistance += lateralVelocityList.get(i);
            }
            return Math.abs(lateralDistance) / x;
        }
    }

    public class FirePowerChooser {

        public final double BULLET_POWER = 2;

        public double firePower() {
            double robotEnergy = battleInfo.botEnergy;
            double neededAmountToKill;
            if (battleInfo.botEnergy <= 4) {
                neededAmountToKill = battleInfo.enemyEnergy / 4;
            } else {
                neededAmountToKill = (battleInfo.enemyEnergy + 2) / 6;
            }
            double energySaver = 1 + 1 / (Math.max(battleInfo.enemyEnergy - robotEnergy, 0) * 0.1 + 1);
            if (TC) {
                return Math.max(Math.min(robotEnergy, 3), 0.1);
            }
            return Math.max(Math.min(neededAmountToKill, Math.min(robotEnergy / 20, Math.min(1200 / battleInfo.distance, Math.max(BULLET_POWER, Math.min(500 / battleInfo.distance, 3))))), 0.1);
        }
    }

    public void onPaint(Graphics2D g) {
        g.setColor(Color.BLUE);
        if (lastWave != null) {
            if (lastWave.distanceTraveled < lastWave.waveVelocity * 3) {
                double[] bins = GunUtils.normalizeBinValues(lastWave.bins);
                for (int i = 0; i < BINS; i++) {
                    double a = lastWave.absBearing + (i - BINS / 2) * 1.0 / (BINS / 2) * lastWave.maxEscapeAngle * lastWave.lateralDirection;
                    Point2D.Double pos2 = GunUtils.project(lastWave.fireLocation, a, bins[i] * 100);
                    g.drawLine((int) lastWave.fireLocation.x,
                            (int) lastWave.fireLocation.y,
                            (int) pos2.x,
                            (int) pos2.y);
                }
            }
        }

        for (int i = 0; i < Math.min(fireAngles.size(), 1); i++) {
            Point2D.Double fireLoc = fireLocations.get(i);
            double maxDeviation = Math.toDegrees(maxDev.get(i));
            Point2D.Double projection = GunUtils.project(fireLoc, fireAngles.get(i), 1000);
            g.setColor(new Color(255, 0, 0, 100));
            g.drawLine((int) fireLoc.x,
                    (int) fireLoc.y,
                    (int) projection.x,
                    (int) projection.y);
            g.setColor(new Color(255, 0, 0, 30));
            g.fillArc((int) fireLoc.x - 1000,
                    (int) fireLoc.y - 1000,
                    2000,
                    2000,
                    (int) (Math.toDegrees(fireAngles.get(i)) - maxDeviation),
                    (int) Math.ceil(maxDeviation * 2));
        }

        g.setColor(Color.WHITE);
        Point2D.Double[] escapePositions = {WSMEAC.getEscapePosition(-1), WSMEAC.getEscapePosition(1)};
        for (int i = 0; i < escapePositions.length; i++) {
            Point2D.Double position = escapePositions[i];
            g.drawRect((int) position.x - 18, (int) position.y - 18, 36, 36);
        }
    }

    public void onBulletHitBullet(BulletHitBulletEvent e) {
        bulletHitBullet++;
    }

    public void onBulletHit(BulletHitEvent e) {
        bulletsHit++;
    }

    public void onRoundEnded(RoundEndedEvent e) {
        for (int i = 0; i < PREDICTORS.size(); i++) {
            System.out.println(PREDICTORS.get(i).getName() + " Score: " + SCORES.get(i));
        }
    }
}
